/* Copyright (c) 2006-2008, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.glimmerblock.api.filter;

import java.net.URL;
import java.util.Map;
import dk.vitality.glimmerblock.api.result.Headers;
import dk.vitality.glimmerblock.api.result.ResultData;
import dk.vitality.glimmerblock.api.util.Logger;

/**
 * The request specific data passed to a BlockerFilter.
 * It contains the requested URL and the client HTTP headers.
 */
abstract public class BlockerFilterData implements Logger
{
	/**
	 * The HTTP method, e.g. "GET".
	 */
	public final String requestMethod;

	/**
	 * URL of the requested content.
	 */
	public final URL requestUrl;

	/**
	 * The clients HTTP headers (e.g. "User-Agent").
	 */
	public final Headers requestHeaders;

	/**
	 * If debugging is enabled. Use the {@link #ln} method to log debug information if this variable is true.
	 * The filter should normally be silent if the flag is false.
	 */
	public final boolean debugFlag;

	/**
	 * The logger for this request.
	 */
	protected final Logger parentLogger;

	protected BlockerFilterData(Logger parentLogger,
	                            String requestMethod,
	                            URL requestUrl,
	                            Headers requestHeaders,
	                            boolean debugFlag)
	{
		this.parentLogger = parentLogger;
		this.requestMethod = requestMethod;
		this.requestUrl = requestUrl;
		this.requestHeaders = requestHeaders;
		this.debugFlag = debugFlag;
	}

	/**
	 * See {@link Logger#ln}
	 */
	public void ln(String s)
	{
		parentLogger.ln(s);
	}

	/**
	 * Send simple response to the client.
	 */
	abstract public void send(ResultData rd) throws Exception;

	/**
	 * Stream response from server.
	 *
	 * @param attachmentFilename If non-null, an "Content-disposition: attachment" header is added with the
	 *                           specified filename.
	 * @param contentTypeMap     Used to change content-type. E.g. youtube sends mp4 as text/html.
	 *                           The content-type which the server sends is looked up in the map,
	 *                           and if a match is found, this content-type
	 *                           is sent to the client instead of the content-type
	 *                           specified by the server.
	 * @param followRedirects    Should redirects be handled internally (true) or should they be
	 *                           forwarded to the client (false).
	 */
	abstract public void stream(String method,
	                            URL url,
	                            String attachmentFilename,
	                            Map<String, String> contentTypeMap, boolean followRedirects) throws Exception;
}