/* Copyright (c) 2006-2008, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.glimmerblock.api.result;

import java.net.HttpURLConnection;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Immutable HTTP headers (and status). It is used for both the request and the response headers.
 * <p/>
 * {@link HeadersBuilder} is to Headers as StringBuilder is to String.
 */
public class Headers
{
	public final int status;
	public final String statusMessage;
	public final String contentType, contentEncoding;

	/**
	 * Maps from lower-case header name to ordered list of values.
	 * Does not contain Content-Type, Content-Encoding, Content-Length and a few other special headers.
	 */
	public final Map<String, List<String>> headerMap;

	/**
	 * Initializes with status = 200 OK and content-type "text/plain" and no content-encoding.
	 */
	public Headers()
	{
		this(HttpURLConnection.HTTP_OK, "OK", "text/plain; charset=UTF-8", null, null);
	}

	/**
	 * Initializes with the specified status and message.
	 */
	public Headers(int status, String statusMessage)
	{
		this(status, statusMessage, "text/plain; charset=UTF-8", null, null);
	}

	/**
	 * Initializes with the specified content-type,
	 * status = 200 OK and no content-encoding.
	 */
	public Headers(String contentType)
	{
		this(HttpURLConnection.HTTP_OK, "OK", contentType, null, null);
	}

	/**
	 * @param status          HTTP status (e.g. 200)
	 * @param statusMessage   HTTP status message (e.g. "OK")
	 * @param contentType     Content-type header (e.g. "text/html")
	 * @param contentEncoding Content-Encoding header (e.g. "gzip")
	 * @param headerMap       A deep copy is made. Maps from lower-case header name to ordered list of values.
	 */
	public Headers(int status, String statusMessage,
	               String contentType, String contentEncoding,
	               Map<String, List<String>> headerMap)
	{
		this.status = status;
		this.statusMessage = statusMessage;
		this.contentType = contentType;
		this.contentEncoding = contentEncoding;
		if (headerMap == null || headerMap.isEmpty()) {
			this.headerMap = Collections.emptyMap();
		} else {
			Map<String, List<String>> map = new HashMap<String, List<String>>(headerMap.size());
			for (String key : headerMap.keySet()) {
				map.put(key, Collections.unmodifiableList(new ArrayList<String>(headerMap.get(key))));
			}
			this.headerMap = Collections.unmodifiableMap(map);
		}
	}

	/**
	 * Compares content-type to the specified value by ignoring charset and other sub-parameters.
	 * This makes "text/plain" equal to "text/plain; charset=utf-8"
	 */
	public boolean isContentType(String ct)
	{
		if (contentType == null)
			return false;
		int idx = contentType.indexOf(';');
		String s = (idx > 0) ? contentType.substring(0, idx).trim() : contentType;
		return s.equals(ct);
	}

	/**
	 * Transforms a lowercase name into the proper mixedcase name.
	 * "content-type" is returned as "Content-Type".
	 *
	 * @param nameLC The <b>lowercase</b> header name.
	 */
	public static String getMixedCaseHeaderName(String nameLC)
	{
		String mx = mixedCaseCache.get(nameLC);
		if (mx != null)
			return mx;
		StringBuilder sb = new StringBuilder(nameLC.length());
		boolean uc = true;
		for (int i = 0; i < nameLC.length(); i++) {
			char ch = nameLC.charAt(i);
			if (uc)
				ch = Character.toUpperCase(ch);
			uc = (ch == '-');
			sb.append(ch);
		}
		mx = sb.toString();
		mixedCaseCache.put(nameLC, mx);
		return mx;
	}

	private static final Map<String, String> mixedCaseCache = new ConcurrentHashMap<String, String>();
}
