/* Copyright (c) 2006-2008, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.glimmerblock.api.result;

import java.io.UnsupportedEncodingException;
import java.net.HttpURLConnection;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import com.sun.grizzly.util.buf.ByteChunk;

/**
 * Immutable structure holding response headers and content.
 */
public class ResultData
{
	/**
	 * The HTTP headers.
	 */
	public final Headers headers;

	/**
	 * The content.
	 */
	protected byte[] content;

	/**
	 * Create without content.
	 */
	public ResultData(Headers headers)
	{
		if (headers == null)
			throw new NullPointerException("headers is null");
		this.headers = headers;
		this.content = null;
	}

	/**
	 * Create with the specified content.
	 *
	 * @param content No copy is made, so the byte array MUST NOT BE MODIFIED by the creator.
	 */
	public ResultData(Headers headers, byte[] content)
	{
		if (headers == null)
			throw new NullPointerException("headers is null");
		this.headers = headers;
		this.content = content;
	}

	/**
	 * @param headers MUST indicate UTF-8 in content-type
	 */
	public ResultData(Headers headers, String content) throws UnsupportedEncodingException
	{
		if (headers == null)
			throw new NullPointerException("headers is null");
		this.headers = headers;
		this.content = content.getBytes("UTF-8");
	}

	/**
	 * Get the content as a string by adhering to the charset specified in the Content-Type header.
	 */
	public String getContentAsText() throws Exception
	{
		Matcher m = charsetRE.matcher(headers.contentType);
		String charset = "ISO-8859-1";
		if (m.find())
			charset = m.group(1);
		if (charset.equals("utf-8lias"))
			charset = "utf-8"; // dilbert.com
		return new String(content, charset);
	}

	static private final Pattern charsetRE = Pattern.compile("\\bcharset=([-A-Za-z0-9]+)", Pattern.CASE_INSENSITIVE);

	/**
	 * Get the content as a Grizzly {@link ByteChunk}
	 *
	 * @return null if it has no content.
	 */
	public ByteChunk getContentAsByteChunk()
	{
		if (content == null || content.length == 0)
			return null;
		ByteChunk chunk = new ByteChunk();
		chunk.setBytes(content, 0, content.length);
		return chunk;
	}

	/**
	 * Simple 404 response.
	 */
	public final static ResultData NOT_FOUND;

	static {
		HeadersBuilder hb = new HeadersBuilder(HttpURLConnection.HTTP_NOT_FOUND, "GlimmerBlocked");
		hb.add("Cache-control", "no-cache, no-store");
		NOT_FOUND = new ResultData(hb.toHeaders());
	}

	/**
	 * GIF with light-gray checkerboard pattern
	 */
	public final static ResultData BLOCKER_GIF;

	static {
		HeadersBuilder hb = new HeadersBuilder();
		hb.add("Content-type", "image/gif");
		hb.add("Cache-control", "no-cache, no-store");
		int[] x = {
				0x47, 0x49, 0x46, 0x38, 0x39, 0x61, 0x04, 0x00, 0x04, 0x00, 0x80, 0x00, 0x00, 0xc8, 0xc8, 0xc8,
				0xff, 0xff, 0xff, 0x21, 0xfe, 0x0e, 0x49, 0x20, 0x77, 0x61, 0x73, 0x20, 0x61, 0x20, 0x62, 0x61,
				0x6e, 0x6e, 0x65, 0x72, 0x00, 0x21, 0xf9, 0x04, 0x01, 0x0a, 0x00, 0x01, 0x00, 0x2c, 0x00, 0x00,
				0x00, 0x00, 0x04, 0x00, 0x04, 0x00, 0x00, 0x02, 0x05, 0x44, 0x7c, 0x67, 0xb8, 0x05, 0x00, 0x3b

		};
		byte[] bytes = new byte[x.length];
		for (int i = 0; i < x.length; i++)
			bytes[i] = (byte)x[i];
		BLOCKER_GIF = new ResultData(hb.toHeaders(), bytes);
	}
}
