/* Copyright (c) 2006-2008, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.glimmerblock.api.util;

import java.io.ByteArrayOutputStream;
import java.io.EOFException;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;

/**
 * Utilities for reading file content and resource data.
 */
public class FileUtils
{
	/**
	 * Get the specified resource.
	 * <p/>
	 * If a file with the specified name exists in the "src" subdirectory,
	 * return the contents of that file (to make developing easier),
	 * otherwise get the resource from the classloader.
	 *
	 * @param path of the resource. Must use slashes for directory separators.
	 */
	public static byte[] getResource(String path) throws IOException
	{
		File f = new File("src", path);
		if (f.exists())
			return readFile(f);
		InputStream is = FileUtils.class.getResourceAsStream(path);
		try {
			ByteArrayOutputStream baos = new ByteArrayOutputStream(is.available());
			int x;
			while ((x = is.read()) >= 0)
				baos.write(x);
			return baos.toByteArray();
		}
		finally {
			is.close();
		}
	}

	/**
	 * Get the contents of the specified file.
	 */
	public static byte[] readFile(File f) throws IOException
	{
		if (!f.isFile())
			throw new IOException("Missing file: " + f.getPath());
		long size = f.length();
		if (size > 1024L * 1024L * 1024L)
			throw new IOException("File too big to read into memory: " + size + " bytes");
		byte[] data = new byte[(int)size];
		FileInputStream fis = new FileInputStream(f);
		try {
			readFully(fis, data);
		}
		finally {
			fis.close();
		}
		return data;
	}

	/**
	 * Read data.length bytes from the stream into the byte array.
	 * Throws EOFException if the stream has less data than needed.
	 */
	public static void readFully(InputStream is, byte[] data) throws IOException
	{
		int offset = 0;
		while (offset < data.length) {
			int num = is.read(data, offset, data.length - offset);
			if (num < 0)
				throw new EOFException();
			offset += num;
		}
	}
}
