/* Copyright (c) 2006-2008, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.glimmerblock.api.util;

/**
 * Utilities for encoding strings (including null) as Javascript string literals.<br>
 * If the parameter is null, the methods returns "null". <br>
 * If the parameter is "Hello, world", the methods returns "'Hello, world'".
 * <p/>
 * The returned value includes the required javascript quotes and is 7-bit ascii clean.
 */
public class JsString
{
	/**
	 * Encode the string as a Javascript string literal.
	 *
	 * @param s The string to encode. Is allowed to be null, which returns a Javascript <code>null</code>.
	 */
	public static String encode(String s)
	{
		return encode(s, '\'');
	}

	/**
	 * Encode the string as a Javascript string literal.
	 *
	 * @param sb The encoded value is appended. The StringBuilder is not reset.
	 * @param s  The string to encode. Is allowed to be null, which returns a Javascript <code>null</code>.
	 */
	public static void encode(StringBuilder sb, String s)
	{
		encode(sb, s, '\'');
	}

	/**
	 * Encode the string as a Javascript string literal.
	 *
	 * @param quote Specifies the quoting character used. MUST BE either single quote or double quote character.
	 * @param s     The string to encode. Is allowed to be null, which returns a Javascript <code>null</code>.
	 */
	public static String encode(String s, char quote)
	{
		if (s == null)
			return "null";
		if (s.length() == 0) {
			if (quote == '\'')
				return "''";
			else if (quote == '"')
				return "\"\"";
		}
		StringBuilder sb = new StringBuilder(s.length() + 1);
		encode(sb, s, quote);
		return sb.toString();
	}

	/**
	 * Encode the string as a Javascript string literal.
	 *
	 * @param sb    The encoded value is appended. The StringBuilder is not reset.
	 * @param s     The string to encode. Is allowed to be null, which returns a Javascript <code>null</code>.
	 * @param quote Specifies the quoting character used. MUST BE either single quote or double quote character.
	 */
	public static void encode(StringBuilder sb, String s, char quote)
	{
		if (s == null) {
			sb.append("null");
			return;
		}
		int len = s.length();
		sb.ensureCapacity(len + 2);
		sb.append(quote);
		for (int i = 0; i < len; i++) {
			char ch = s.charAt(i);
			switch (ch) {
				case '\r':
				case '\n':
					sb.append("\\n");
					break;
				case '\t':
					sb.append("\\t");
					break;
				case '"':
					sb.append("\\x22");
					break;
				case '\'':
					sb.append("\\x27");
					break;
				case '<':
					sb.append("\\x3c");
					break;
				case '>':
					sb.append("\\x3e");
					break;
				default:
					if (ch > 255) {
						sb.append("\\u");
						sb.append(String.format("%04x", (int)ch));
					} else if (ch < ' ' || ch >= 127) {
						sb.append("\\x");
						sb.append(String.format("%02x", (int)ch));
					} else {
						sb.append(ch);
					}
			}
		}
		sb.append(quote);
	}
}