/* Copyright (c) 2006-2008, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.glimmerblock.proxy;

import java.io.IOException;
import com.sun.grizzly.http.KeepAliveCountManager;
import com.sun.grizzly.http.KeepAliveStats;
import com.sun.grizzly.http.SelectorThread;
import com.sun.grizzly.tcp.ActionCode;
import com.sun.grizzly.tcp.Adapter;
import com.sun.grizzly.tcp.OutputBuffer;
import com.sun.grizzly.tcp.Request;
import com.sun.grizzly.tcp.Response;
import com.sun.grizzly.util.buf.ByteChunk;
import com.sun.grizzly.util.buf.MessageBytes;
import dk.vitality.glimmerblock.proxy.mgr.GlimmerManager;
import dk.vitality.glimmerblock.proxy.request.RequestHandler;
import static dk.vitality.glimmerblock.proxy.util.ExceptionUtils.getExMsg;

/**
 * Main class. If you want to add your own filters, subclass this class and override createManager()
 */
public class GlimmerBlock
{
	public static void main(String[] args)
	{
		try {
			System.setProperty("http.maxConnections", "100");
			GlimmerBlock gb = new GlimmerBlock();
			gb.main();
		}
		catch (Exception ex) {
			System.err.println("GlimmerBlock got exception in main thread: " + getExMsg(ex));
			ex.printStackTrace();
			System.exit(1);
		}
	}

	public boolean debugFlag = false;
	public boolean debugVerbose = false;
	public static final boolean debugAdapter = false;

	protected MySelectorThread selectorThread;

	protected GlimmerManager mgr;

	/**
	 * Initialize and run the server.
	 *
	 * @throws Exception
	 */
	public void main() throws Exception
	{
		try {
			createManager();
			createSelectorThread();
			runGrizzlyServer();
		}
		finally {
			if (mgr != null)
				mgr.ln("quits");
		}
	}

	/**
	 * Initializes the filters.
	 */
	protected void createManager() throws Exception
	{
		mgr = new GlimmerManager();
		mgr.init();
	}

	/**
	 * Creates the Grizzly instance.
	 */
	protected void createSelectorThread() throws Exception
	{
		selectorThread = new MySelectorThread();
		if (mgr.listenAddress != null) {
			mgr.ln("Listens to: " + mgr.listenAddress + ":" + mgr.listenPort);
			selectorThread.setAddress(mgr.listenAddress);
		} else {
			mgr.ln("Listens to: *:" + mgr.listenPort);
		}
		selectorThread.setPort(mgr.listenPort);
		selectorThread.setAdapter(new ProxyAdapter());
		selectorThread.setCompression("off"); // don't let proxy compress stuff that's already compressed.
		selectorThread.setDisplayConfiguration(true);
		selectorThread.setFileCacheIsEnabled(false);
		selectorThread.setLargeFileCacheEnabled(false);
		selectorThread.setBufferResponse(false);
		selectorThread.setKeepAliveTimeoutInSeconds(35);
		selectorThread.setMaxThreads(50);
		//selectorThread.setBufferSize(32768);
		selectorThread.setMaxKeepAliveRequests(8196);
		//st.setKeepAliveThreadCount(500);
	}

	/**
	 * Runs the grizzly http server. Returns only when the server is shutting down (i.e. never).
	 */
	protected void runGrizzlyServer() throws Exception
	{
		try {
			selectorThread.initEndpoint();
			selectorThread.enableMonitoring();
			mgr.ln("Starts endpoint");
			selectorThread.startEndpoint();
		} catch (Exception ex) {
			mgr.ln("Exception in SelectorThread: " + ex.getClass().getSimpleName() + ": " + ex.getMessage());
		} finally {
			if (selectorThread.isRunning()) {
				mgr.ln("stops endpoint");
				selectorThread.stopEndpoint();
			}
		}
	}

	/**
	 * Process a request.  Subclass if you subclass {@link RequestHandler}.
	 */
	protected void processProxyRequest(Request request, Response response) throws Exception
	{
		RequestHandler handler = new RequestHandler(request, response, mgr, debugFlag, debugVerbose);
		handler.main();
	}

	/**
	 * Handles simple requests for statistics which are not proxied.
	 *
	 * @return true if the URI is a statistics request, and false if it's not (and should be proxied).
	 */
	@SuppressWarnings({"UnusedDeclaration"})
	protected boolean processStatisticsRequest(Request request, Response response, MessageBytes uri) throws Exception
	{
		if (uri.equals("/grizzly/dummy")) {
			sendPlainTextResponse(response, "Hello, world.\n");
			return true;
		}
		if (uri.equals("/grizzly/statistics")) {
			sendPlainTextResponse(response, selectorThread.getStatistics());
			return true;
		}
		if (uri.equals("/grizzly/rrd-statistics")) {
			KeepAliveStats kas = selectorThread.getKeepAliveStats();
			String txt;
			if (kas == null)
				txt = "0\n";
			else
				txt = kas.getCountConnections() + "\n";
			sendPlainTextResponse(response, txt);
			return true;
		}
		return false;
	}

	/**
	 * Request interface to Grizzly. Kind of lowlevel servlet interface.
	 */
	class ProxyAdapter implements Adapter
	{
		public void service(Request request, Response response) throws Exception
		{
			if (debugAdapter)
				System.out.println("ProxyAdapter.enters: " + request.method() + " " + request.unparsedURI());
			try {
				if (processStatisticsRequest(request, response, request.unparsedURI()))
					return;
				request.action(ActionCode.ACTION_REQ_HOST_ADDR_ATTRIBUTE, null);
				processProxyRequest(request, response);
			}
			finally {
				if (debugAdapter)
					System.out.println("ProxyAdapter.leaves: " + request.method() + " " + request.unparsedURI());
			}
		}

		public void afterService(Request request, Response response) throws Exception
		{
			if (debugAdapter)
				mgr.ln("afterService: " + request);
			request.recycle();
			response.recycle();
		}

		public void fireAdapterEvent(String string, Object object)
		{
			if (debugAdapter)
				mgr.ln("fireAdapterEvent: " + string);
		}
	}

	/**
	 * Send a simple text response. Used for statistics reporting, performance test, etc, which
	 * doesn't need UTF-8 characters.
	 *
	 * @param msg Only the ISO-8859-1 subset of characters is allowed.
	 */

	public static void sendPlainTextResponse(Response response, String msg) throws IOException
	{
		byte[] b = msg.getBytes("ISO-8859-1");
		response.setContentType("text/plain; charset=ISO-8859-1");
		response.setContentLength(b.length);
		ByteChunk chunk = new ByteChunk();
		chunk.append(b, 0, b.length);
		OutputBuffer buffer = response.getOutputBuffer();
		buffer.doWrite(chunk, response);
		response.finish();
	}

	/**
	 * Simple subclass of {@link SelectorThread} which provides access to keep-alive statistics.
	 */
	static class MySelectorThread extends SelectorThread
	{
		public String getStatistics()
		{
			KeepAliveCountManager kacm = keepAliveCounter;
			if (kacm == null)
				return "missing keepAliveCounter";
			String s = kacm.getCurrentThreadCount() + " " + kacm.getCurrentThreadsBusy() + "\n";
			KeepAliveStats kas = getKeepAliveStats();
			if (kas != null) {
				s += "KAS:  conns=" + kas.getCountConnections()
						+ ", flushes=" + kas.getCountFlushes()
						+ ", hits=" + kas.getCountHits()
						+ ", refusals=" + kas.getCountRefusals()
						+ ", timeouts=" + kas.getCountTimeouts()
						+ ", maxConns=" + kas.getMaxConnections()
						+ ", secTimeouts=" + kas.getSecondsTimeouts() + "\n";
			}
			return s;
		}
	}
}
