/* Copyright (c) 2006-2008, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.glimmerblock.proxy.filter;

import java.io.IOException;
import java.util.regex.Pattern;
import static java.util.regex.Pattern.CASE_INSENSITIVE;
import static java.util.regex.Pattern.DOTALL;
import static java.util.regex.Pattern.compile;
import dk.vitality.glimmerblock.api.filter.ContentFilter;
import dk.vitality.glimmerblock.api.filter.ContentFilterData;
import dk.vitality.glimmerblock.api.filter.ContentTextHtmlFilter;
import dk.vitality.glimmerblock.api.util.FileUtils;
import dk.vitality.glimmerblock.api.util.JsString;

/**
 * Utility base class for creating content-modifying filters.
 */
public class AbstractContentFilterAdder
{
	public final DomainTreeSet filterSet;

	public AbstractContentFilterAdder(DomainTreeSet filterSet)
	{
		this.filterSet = filterSet;
	}

	/**
	 * Convenience method to add the specified filter to the text/html filterSet.
	 *
	 * @param hostPath See {@link DomainTree#add}
	 */
	protected void addHtml(String hostPath, ContentFilter filter)
	{
		filterSet.textHtml.add(hostPath, filter);
	}

	/**
	 * Convenience method to add the specified filter to the text/xml filterSet.
	 *
	 * @param hostPath See {@link DomainTree#add}
	 */
	protected void addXml(String hostPath, ContentFilter filter)
	{
		filterSet.textXml.add(hostPath, filter);
	}


	/**
	 * Returns a &lt;script&gt;...&gt;/script&gt; string with the specified javascript content.
	 */
	public static String scriptElement(String js)
	{
		return "<script type='text/javascript'> // GlimmerBlock script\n" + js.trim() + "\n</script>\n";
	}

	/**
	 * Returns a javascript snippet which sets the documents title to the specified string.
	 */
	public static String setTitleJSE(String title)
	{
		return scriptElement("document.title = " + jsEncode(title) + ";");
	}

	/**
	 * Get the contents of the external script file.
	 */
	public static String scriptFileContents(String path) throws IOException
	{
		byte[] bytes = FileUtils.getResource(path);
		return scriptElement(new String(bytes, "UTF-8"));
	}

	/**
	 * Convenience method for calling {@link JsString#encode}
	 */
	public static String jsEncode(String s)
	{
		return JsString.encode(s);
	}

	/**
	 * Simple filter which zaps all &lt;script&gt; tags in the html and thus disable most javascripts on the page.
	 */
	public static class ScriptZapper extends ContentTextHtmlFilter
	{
		protected String filterHtml(String html, ContentFilterData data) throws Exception
		{
			return scriptRE.matcher(html).replaceAll("<!-- GB zapped script -->");
		}

		public static final Pattern scriptRE = compile("<script.*?</script>", DOTALL);
	}

	/**
	 * Simple filter which zaps all &lt;iframe&gt;s in the html and thus disable many ads as some sites
	 * always displays ads using iframes.
	 */
	public static class IFrameFilter extends ContentTextHtmlFilter
	{
		protected String filterHtml(String html, ContentFilterData data) throws Exception
		{
			return iframeRE.matcher(html).replaceAll("<!-- GB zapped iframe -->");
		}

		public static final Pattern iframeRE = compile("<iframe.*?</iframe\\s*>", DOTALL | CASE_INSENSITIVE);
	}

}
