/* Copyright (c) 2006-2008, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.glimmerblock.proxy.filter;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import dk.vitality.glimmerblock.api.util.Logger;

/**
 * For fast matching of hostnames or domain-wildcards to hostnames.
 * The hostnames are stored using a tree by splitting it at the dots,
 * and using the root domain as the tree root.
 */
public class DomainTree<FILTER>
{
	public final Logger logger;
	protected DomainNode<FILTER> root = new DomainNode<FILTER>();

	public DomainTree(Logger logger)
	{
		this.logger = logger;
	}

	/**
	 * @return Never null. Empty list if none matches.
	 *         The list is sorted such that more specific filters
	 *         comes before less specific filters.
	 *         The filters are sorted by domain name, then initialPath.
	 */
	public List<FILTER> getFilters(String host, String path)
	{
		List<FILTER> list = new ArrayList<FILTER>(1);
		root.getSubtreeMatches(host, path, list);
		return list;
	}

	/**
	 * @param hostPath Must always contain a host. Can optionally contain an initial url-path which the request
	 *                 must match.<br>
	 *                 If the host is prepended by a dot, it matches the full subdomain.
	 *                 E.g. "adtech.de" matches only the host "adtech.de",
	 *                 but ".adtech.de" matches "adtech.de", "www.adtech.de", "adserver.adtech.de", etc.
	 * @param filter   The filter to add.
	 */
	public synchronized void add(String hostPath, FILTER filter)
	{
		int slash = hostPath.indexOf('/');
		boolean wildcard = hostPath.startsWith(".");
		String initialPath = null;
		String s;
		if (slash >= 0) {
			initialPath = hostPath.substring(slash);
			s = hostPath.substring(wildcard ? 1 : 0, slash);
		} else {
			s = hostPath.substring(wildcard ? 1 : 0);
		}
		DomainNode<FILTER> n = root;
		while (true) {
			if (s.length() == 0) {
				n.add(new FilterInfo<FILTER>(filter, wildcard, initialPath));
				return;
			}
			int idx = s.lastIndexOf('.');
			String key = (idx < 0) ? s : s.substring(idx + 1);
			if (n.subNodes == null)
				n.subNodes = new ConcurrentHashMap<String, DomainNode<FILTER>>();
			DomainNode<FILTER> sub = n.subNodes.get(key);
			if (sub == null) {
				sub = new DomainNode<FILTER>();
				n.subNodes.put(key, sub);
			}
			n = sub;
			s = idx < 0 ? "" : s.substring(0, idx);
		}
	}

	/**
	 * A Node in the tree. Contains a map of sub-nodes and the list of filters for this node.
	 */
	protected static class DomainNode<FILTER>
	{
		Map<String, DomainNode<FILTER>> subNodes;
		List<FilterInfo<FILTER>> filterList;

		/**
		 * Adds the filters which matches the subtree. Deep-first search so the more specific matches
		 * are in front of the list.
		 *
		 * @param subhost The hostname seen from this node.  "adserver.adtech.de" becomes "adserver.adtech",
		 *                then "adserver" and finally "" when the method recurses.
		 */
		void getSubtreeMatches(String subhost, String path, List<FILTER> list)
		{
			if (subhost.length() == 0) {
				getNodeMatches(list, false, path);
				return;
			}
			if (subNodes != null) {
				int idx = subhost.lastIndexOf('.');
				String key = (idx < 0) ? subhost : subhost.substring(idx + 1);
				DomainNode<FILTER> n = subNodes.get(key);
				if (n != null)
					n.getSubtreeMatches(idx < 0 ? "" : subhost.substring(0, idx), path, list);
			}
			getNodeMatches(list, true, path);
		}

		/**
		 * Add the filters for this node which matches the path.
		 */
		void getNodeMatches(List<FILTER> list, boolean needsWildcard, String path)
		{
			if (filterList == null)
				return;
			for (FilterInfo<FILTER> fi : filterList) {
				if ((fi.wildcard || !needsWildcard) && (fi.initialPath == null || path.startsWith(fi.initialPath)))
					list.add(fi.filter);
			}
		}

		/**
		 * Adds the filter to this node. Ensures the filterList is sorted such that filters
		 * with a longer initialPath comes before those with a shorter/none.
		 */
		void add(FilterInfo<FILTER> info)
		{
			if (filterList == null)
				filterList = Collections.synchronizedList(new ArrayList<FilterInfo<FILTER>>(1));
			if (info.initialPath != null) {
				for (int i = 0; i < filterList.size(); i++) {
					FilterInfo<FILTER> x = filterList.get(i);
					if (x.initialPath == null || x.initialPath.length() > info.initialPath.length()) {
						filterList.add(i, info);
						return;
					}
				}
			}
			filterList.add(info);
		}
	}

	/**
	 * Matching-information for a specific filter.
	 */
	protected static class FilterInfo<FILTER>
	{
		final FILTER filter;
		/**
		 * Null if it matches all paths. Non-null if the filter only matches a sub-path.
		 */
		final String initialPath;
		/**
		 * Matches the filter all sub-domains (true) or only this specific host (false).
		 */
		final boolean wildcard;

		FilterInfo(FILTER filter, boolean wildcard, String initialPath)
		{
			this.filter = filter;
			this.initialPath = initialPath;
			this.wildcard = wildcard;
		}
	}
}
