/* Copyright (c) 2006-2008, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.glimmerblock.proxy.mgr;

import java.net.InetAddress;
import java.util.HashSet;
import java.util.Set;
import dk.vitality.glimmerblock.api.filter.BlockerFilterData;
import dk.vitality.glimmerblock.api.util.Logger;
import dk.vitality.glimmerblock.proxy.filter.DomainTreeSet;
import dk.vitality.glimmerblock.proxy.standardFilters.StandardBlockerFilters;
import dk.vitality.glimmerblock.proxy.standardFilters.StandardContentFilters;
import dk.vitality.glimmerblock.proxy.util.LoggerImpl;

/**
 * Intended for preferences and more global customization without requiring you to subclass
 * more internal classes.
 */
public class GlimmerManager implements Logger
{
	public DomainTreeSet filterSet;

	/**
	 * If null, listens to all local adresses. If non-null, binds to the specified interface.
	 */
	public InetAddress listenAddress;
	/**
	 * The port that Grizzly should listen to.
	 */
	public int listenPort;
	/**
	 * The log handler.
	 */
	protected Logger logger;

	/**
	 * The allowed HTTP methods (GET, POST, ...)
	 */
	protected Set<String> allowedMethods;

	/**
	 * Is the filters enabled, or are the requests always sent unmodified.
	 */
	public boolean filtersEnabled = true;

	/**
	 * Is the management handler enabled, or do we ignore requests to enable/disable filters etc.
	 */
	public boolean managementEnabled = true;

	/**
	 * The pseudo-domain used for management requests.
	 * Beware that it is hardcoded in *.js files.
	 */
	public static final String managementHost = "gb.test";

	/**
	 * Initialize the manager. Must be called after instance creation.
	 */
	public void init() throws Exception
	{
		createLogger();
		createAllowedMethods();
		createListenAddress();
		createFilterSet();
	}

	/**
	 * Is the HTTP method allowed. Called by {@link dk.vitality.glimmerblock.proxy.request.RequestHandler}
	 */
	public boolean isMethodAllowed(String method)
	{
		return allowedMethods.contains(method);
	}

	/**
	 * Handle a management request such as enable/disable filters, special download handling.
	 */
	public void handleManagementRequest(BlockerFilterData data) throws Exception
	{
		new ManagementFilter(this, data).handleRequest();
	}

	/**
	 * Create the logger instance. Override the method if you want to use another logger implementation.
	 */
	protected void createLogger() throws Exception
	{
		logger = new LoggerImpl();
	}

	/**
	 * Initialize the supported HTTP method. Override the method if you want to modify the set.
	 */
	protected void createAllowedMethods() throws Exception
	{
		allowedMethods = new HashSet<String>();
		for (String s : "GET,HEAD,POST,OPTIONS,PROPFIND,PUT,LOCK,UNLOCK,DELETE,MOVE,COPY".split(","))
			allowedMethods.add(s);
	}

	/**
	 * Initialize which IP address and port number that the proxy listens to.
	 */
	protected void createListenAddress()
	{
		this.listenAddress = null;//InetAddress.getByName("127.0.0.1");
		this.listenPort = 8228;
	}

	/**
	 * Create the filterSet instance.
	 */
	protected void createFilterSet()
	{
		this.filterSet = new DomainTreeSet(logger);
		initializeFilterSet();
	}

	/**
	 * Adds the standard set of filters. Override if you want to add more filters or want another set.
	 */
	protected void initializeFilterSet()
	{
		new StandardBlockerFilters(filterSet);
		new StandardContentFilters(filterSet);
	}

	/**
	 * See {@link Logger#ln}
	 */
	public void ln(String s)
	{
		logger.ln(s);
	}
}
