/* Copyright (c) 2006-2008, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.glimmerblock.proxy.mgr;

import java.io.ByteArrayOutputStream;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;
import dk.vitality.glimmerblock.api.filter.BlockerFilterData;
import dk.vitality.glimmerblock.api.result.Headers;
import dk.vitality.glimmerblock.api.result.ResultData;
import dk.vitality.glimmerblock.api.util.FileUtils;
import dk.vitality.glimmerblock.api.util.Logger;
import dk.vitality.glimmerblock.proxy.util.UrlParams;

/**
 * Servlet for handling special requests such as enabling/disabling filters, youtube downloads etc.
 */
public class ManagementFilter implements Logger
{
	public final GlimmerManager mgr;
	public final BlockerFilterData data;
	public final UrlParams sp;

	public ManagementFilter(GlimmerManager mgr, BlockerFilterData data) throws Exception
	{
		this.mgr = mgr;
		this.data = data;
		this.sp = new UrlParams();
		this.sp.addFromQueryString(data.requestUrl.getQuery());
	}

	/**
	 * Handle the management request. Called by {@link dk.vitality.glimmerblock.proxy.request.RequestHandler}
	 */
	public void handleRequest() throws Exception
	{
		data.ln(null);
		String path = data.requestUrl.getPath();
		if (path == null || path.length() == 0)
			path = "/";
		dispatch(path);
	}

	/**
	 * Handle the request.
	 *
	 * @param path The request path. Never null or empty string.
	 */
	protected void dispatch(String path) throws Exception
	{
		if (path.equals("/toggle")) {
			handleToggle();
			return;
		}
		if (path.equals("/download")) {
			handleDownload();
			return;
		}
		Headers h = new Headers(HttpURLConnection.HTTP_NOT_FOUND, "Unknown GB action: " + path);
		data.send(new ResultData(h, h.statusMessage + "\n"));
	}

	/**
	 * Toggles the flag for if filters are enabled/disabled.
	 */
	protected void handleToggle() throws Exception
	{
		mgr.filtersEnabled ^= true;
		if (mgr.filtersEnabled)
			sendMgrJS("_GlimBlock_.showMsg('ENABLED', 'green');\n");
		else
			sendMgrJS("_GlimBlock_.showMsg('disabled', 'red');\n");
	}

	/**
	 * Handles special downloads where attachment title are forced to
	 * something different from the originating servers HTTP headers.
	 */
	protected void handleDownload() throws Exception
	{
		String title = sp.reqString("title");
		String fn = cleanFilename(title + ".mpg");
		URL url = new URL(sp.reqString("url"));
		ln("Simple download of " + url);
		ln("  saved as: " + fn);
		handleSimpleDownload(url, fn, "text/html", "video/mpeg");
	}

	/**
	 * Utility method for simple downloads.
	 *
	 * @param url                The URL of the document.
	 * @param attachmentFilename The local filename which the browser saves the document as.
	 * @param contentTypeMap     A list of string-pairs for mapping content-types as many servers send errornous
	 *                           content-types for e.g. videos.
	 */
	protected void handleSimpleDownload(URL url, String attachmentFilename, String... contentTypeMap) throws Exception
	{
		Map<String, String> ctMap = new HashMap<String, String>();
		for (int i = 0; i < contentTypeMap.length - 1; i += 2)
			ctMap.put(contentTypeMap[i], contentTypeMap[i + 1]);
		data.stream("GET", url, attachmentFilename, ctMap, true);
	}

	/**
	 * Sends the management script appended by the specified string.
	 *
	 * @param extra Javascript which is sent in addition to the management javascript file.
	 */
	protected void sendMgrJS(String extra) throws Exception
	{
		byte[] script = FileUtils.getResource("dk/vitality/glimmerblock/proxy/mgr/GlimmerManager.js");
		ByteArrayOutputStream baos = new ByteArrayOutputStream(script.length + extra.length() + 32);
		baos.write(script);
		baos.write(extra.getBytes("UTF-8"));
		Headers h = new Headers("application/x-javascript; charset=UTF-8");
		data.send(new ResultData(h, baos.toByteArray()));
	}

	/**
	 * Clean a string so it is suitable as a safe client filename: ensure it
	 * <ul>
	 * <li>doesn't contain slashes</li>
	 * <li>start with a dot</li>
	 * <li>contains ascii control characters</li>
	 * <li>is the empty string</li>
	 * </ul>
	 */
	public static String cleanFilename(String title)
	{
		StringBuilder sb = new StringBuilder(title.length());
		for (int i = 0; i < title.length(); i++) {
			char ch = title.charAt(i);
			if (ch < 32)
				continue;
			if (ch == '.' && sb.length() == 0)
				continue;
			if (ch == '/')
				continue;
			sb.append(ch);
		}
		if (sb.length() == 0)
			sb.append("without-name");
		return sb.toString();
	}

	/**
	 * See {@link Logger#ln}
	 */
	public void ln(String s)
	{
		data.ln(s);
	}
}
