/* Copyright (c) 2006-2008, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.glimmerblock.proxy.request;

import java.io.IOException;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import com.sun.grizzly.tcp.Request;
import com.sun.grizzly.tcp.Response;
import com.sun.grizzly.util.buf.ByteChunk;
import dk.vitality.glimmerblock.api.filter.BlockerFilter;
import dk.vitality.glimmerblock.api.filter.ContentFilter;
import dk.vitality.glimmerblock.api.filter.ContentFilterData;
import dk.vitality.glimmerblock.api.result.Headers;
import dk.vitality.glimmerblock.api.result.ResultData;
import dk.vitality.glimmerblock.proxy.filter.DomainTreeSet;
import dk.vitality.glimmerblock.proxy.mgr.GlimmerManager;
import dk.vitality.glimmerblock.proxy.urlExec.HttpException;
import dk.vitality.glimmerblock.proxy.urlExec.UrlStreamer;
import static dk.vitality.glimmerblock.proxy.util.ExceptionUtils.getExMsg;

/**
 * The main class for handling HTTP proxy requests.
 */
public class RequestHandler extends RequestUtils
{
	/**
	 * Preferences etc.
	 */
	public final GlimmerManager mgr;

	/**
	 * Filters.
	 */
	public final DomainTreeSet filterSet;

	/**
	 * Do we allow "Content-Encoding: gzip" from the server?
	 */
	public final boolean allowGzip;

	/**
	 * URL of the requested content.
	 */
	public URL requestURL;

	/**
	 * The path part of {@link #requestURL}. Contains "/" instead of empty string if no path was specified.
	 */
	public String requestPath;

	public RequestHandler(Request request,
	                      Response response,
	                      GlimmerManager mgr,
	                      boolean debugFlag,
	                      boolean debugVerbose) throws Exception
	{
		super(request, response, mgr, debugFlag, debugVerbose);
		this.mgr = mgr;
		this.filterSet = mgr.filterSet;
		if (this.debugFlag)
			ln(null);
		if (this.debugVerbose)
			ln("RequestHandler: " + Thread.currentThread().getName());
		String acceptEncoding = request.getHeader("Accept-Encoding");
		allowGzip = acceptEncoding != null && acceptEncoding.contains("gzip");
	}

	/**
	 * The main method for executing the request.
	 */
	public void main() throws Exception
	{
		try {
			checkMethodIsAllowed();
			parseUri();
			execute();
		}
		catch (ClientAbortedException ex) {
			if (debugFlag)
				ln("Got ClientAbortedException: " + ex.getMessage());
		}
		catch (HttpException ex) {
			if (ex.headers == null) {
				ln("Got HttpException with null headers: " + ex.getMessage());
				ex.printStackTrace();
			} else {
				ResultData rd = new ResultData(ex.headers, ex.headers.statusMessage + "\n");
				sendResultData(rd);
			}
		}
		catch (Exception ex) {
			ln("GOT exception: " + ex.getMessage());
			ex.printStackTrace();
		}
		finally {
			response.finish();
			if (debugVerbose)
				ln("All done.");
		}
	}

	/**
	 * Do we allow the requested method, or do we outright deny it (such as CONNECT).
	 */
	protected void checkMethodIsAllowed() throws HttpException
	{
		if (!mgr.isMethodAllowed(requestMethod)) {
			Headers h = new Headers(HttpURLConnection.HTTP_BAD_METHOD, "Method " + requestMethod + " not allowed");
			throw new HttpException(h);
		}
	}

	/**
	 * Parse the request path sent from the client. It must be a full URL as this
	 * is a proxy server.
	 * @return false if an error occured and a response was sent to the client.
	 * @throws IOException
	 */
	protected void parseUri() throws HttpException
	{
		String s = request.unparsedURI().toString();
		try {
			requestURL = new URL(s);
			requestPath = requestURL.getPath();
			if (requestPath == null || requestPath.length() == 0)
				requestPath = "/";
		}
		catch (Exception ex) {
			Headers h = new Headers(HttpURLConnection.HTTP_BAD_REQUEST,
					"Bad request URL (" + getExMsg(ex) + "): " + s);
			throw new HttpException(h, ex);
		}
	}

	/**
	 * Execute the request after method and URL has been validated/decoded.
	 * <p>
	 * Handles blocked URLs.
	 */
	protected void execute() throws Exception
	{
		String host = requestURL.getHost();
		if (mgr.managementEnabled && host.equals(GlimmerManager.managementHost)) {
			// can't use blocker tree as we disable that
			ContentFilterDataImpl fd = new ContentFilterDataImpl(this, null);
			mgr.handleManagementRequest(fd);
			return;
		}
		List<BlockerFilter> blockerList = mgr.filtersEnabled ? filterSet.blockers.getFilters(host, requestPath) : null;
		if (!blockerList.isEmpty()) {
			if (debugFlag)
				ln("Got " + blockerList.size() + " BlockerFilters");
			for (BlockerFilter blocker : blockerList) {
				ContentFilterDataImpl fd = new ContentFilterDataImpl(this, null);
				blocker.process(fd);
				if (fd.hasSentResponse)
					return;
			}
			if (debugFlag)
				ln("None of the BlockFilters blocked.");
		}
		RequestStreamer streamer = new RequestStreamer(this, allowGzip);
		streamer.execute();
	}

	/**
	 * Handles streaming the content and optinally filtering it using the matching content filters.
	 */
	class RequestStreamer extends UrlStreamer
	{
		RequestStreamer(RequestHandler rh, boolean allowGzip) throws Exception
		{
			super(rh, rh.requestMethod, rh.requestURL, rh.requestHeaders, allowGzip, rh);
			this.debugFlag = rh.debugVerbose;
		}

		/**
		 * Adds POST request data.
		 */
		protected void sendRequestBody() throws Exception
		{
			if (!method.equals("POST"))
				return;
			long requestContentLength = request.getContentLengthLong();
			conn.setDoOutput(requestContentLength > 0);
			conn.setDoInput(!isHEAD);
			if (debugVerbose)
				ln("sendRequestBody, requestContentLength = " + requestContentLength);
			if (requestContentLength <= 0)
				return;
			ByteChunk chunk = new ByteChunk();
			long remaining = requestContentLength;
			OutputStream os = conn.getOutputStream();
			while (remaining > 0) {
				chunk.setLimit((int)Math.min(remaining, 65536));
				int num = request.doRead(chunk);
				if (debugVerbose) {
					ln("InputSender: num = " + num + ", chunk.len = " + chunk.getLength()
							+ ", remaining = " + remaining + " / " + requestContentLength);
				}
				if (chunk.getLength() <= 0) {
					// can this ever happend???
					Thread.sleep(50);
					continue;
				}
				if (chunk.getBuffer() == null)
					throw new NullPointerException("chunk.getBuffer() is null");
				byte[] data = chunk.getBuffer();
				int offset = chunk.getOffset();
				if (debugVerbose)
					ln("POST data: '" + new String(data, offset, num, "ISO-8859-1") + "'");
				os.write(data, offset, num);
				remaining -= num;
			}
			os.close();
			if (debugVerbose) {
				ln("InputSender: done sending, " + remaining + " bytes of " +
						requestContentLength + " remaining.");
			}
		}

		/**
		 * Process the response by e.g. filtering it.
		 */
		protected void streamResponseBody() throws Exception
		{
			if (!mgr.filtersEnabled || responseHeaders.status != HttpURLConnection.HTTP_OK) {
				super.streamResponseBody();
				return;
			}
			String ct = responseHeaders.contentType;
			if (ct == null)
				ct = "text/plain";
			int idx = ct.indexOf(';');
			if (idx > 0)
				ct = ct.substring(0, idx).trim();
			List<ContentFilter> filterList;
			if (ct.equals("text/html"))
				filterList = filterSet.textHtml.getFilters(requestUrl.getHost(), requestPath);
			else if (ct.equals("text/xml"))
				filterList = filterSet.textXml.getFilters(requestUrl.getHost(), requestPath);
			else
				filterList = Collections.emptyList();
			if (filterList.isEmpty()) {
				super.streamResponseBody();
				return;
			}
			if (debugFlag)
				ln("Got " + filterList.size() + " ContentFilters");
			readResponseBodyIntoByteArray();
			ResultData rd = new ResultData(responseHeaders, responseContent);
			ContentFilterDataImpl fd = new ContentFilterDataImpl(RequestHandler.this, rd);
			for (ContentFilter filter : filterList) {
				if (debugFlag)
					ln("ContentFilter: " + filter.getClass().getSimpleName());
				filter.process(fd);
				if (fd.hasSentResponse)
					return;
			}
			fd.send(rd);
		}
	}

	/**
	 * Used too for blocker filters, but with null "ResultData responseData"
	 */
	class ContentFilterDataImpl extends ContentFilterData
	{
		/**
		 * Has the filter sent a response to the client?
		 */
		boolean hasSentResponse;

		ContentFilterDataImpl(RequestHandler rh, ResultData responseData)
		{
			super(rh, rh.requestMethod, rh.requestURL, rh.requestHeaders, rh.debugFlag, responseData);
		}

		public void send(ResultData rd) throws Exception
		{
			startResponse();
			sendResultHeaders(rd.headers);
			ByteChunk chunk = rd.getContentAsByteChunk();
			if (chunk != null)
				response.doWrite(chunk);
		}

		public void stream(String method,
		                   URL url,
		                   final String attachmentFilename,
		                   final Map<String, String> contentTypeMap,
		                   final boolean followRedirects) throws Exception
		{
			startResponse();
			UrlStreamer streamer = new MyUrlStreamer(RequestHandler.this,
					method, url, attachmentFilename, contentTypeMap, followRedirects);
			streamer.execute();
		}

		/**
		 * Checks we haven't already sent a response.
		 * Sets the {@link #hasSentResponse} flag, so this is a one-time only.
		 */
		private void startResponse() throws Exception
		{
			if (hasSentResponse)
				throw new Exception("Has already sent response");
			hasSentResponse = true;
		}
	}

	/**
	 * Utility class for {@link ContentFilterDataImpl} for streaming an URL to the client.
	 * Used when a filter wants to stream a different URL from the server, e.g.
	 * YouTube video downloads.<p>
	 * See {@link dk.vitality.glimmerblock.api.filter.BlockerFilterData#stream}
	 */
	private class MyUrlStreamer extends UrlStreamer
	{
		private final String attachmentFilename;
		private final Map<String, String> contentTypeMap;

		MyUrlStreamer(RequestHandler rh,
		              String method,
		              URL url,
		              String attachmentFilename,
		              Map<String, String> contentTypeMap, boolean followRedirects) throws Exception
		{
			super(rh, method, url, rh.requestHeaders, rh.allowGzip, rh);
			this.followRedirects = followRedirects;
			this.attachmentFilename = attachmentFilename;
			this.contentTypeMap = contentTypeMap;
		}

		protected void copyBrowserHeaders() throws Exception
		{
			ln("stream.copyBrowserHeaders: " + attachmentFilename);
			super.copyBrowserHeaders();
			if (attachmentFilename != null)
				response.setHeader("Content-Disposition", "attachment; filename=" + attachmentFilename);
		}

		@SuppressWarnings({"SimplifiableIfStatement"})
		protected boolean shouldCopyBrowserHeader(String nameLC)
		{
			if (nameLC.equals("content-disposition") && attachmentFilename != null)
				return false;
			return basicBrowserHeadersLC.contains(nameLC);
		}

		protected void sendResultHeaders() throws Exception
		{
			if (responseHeaders.contentType == null || contentTypeMap == null) {
				super.sendResultHeaders();
				return;
			}
			String ct = responseHeaders.contentType;
			int idx = ct.indexOf(';');
			if (idx > 0)
				ct = ct.substring(0, idx).trim();
			String ct2 = contentTypeMap.get(ct.toLowerCase());
			if (ct2 == null) {
				rinfo.sendResultHeaders(responseHeaders);
				return;
			}
			if (idx > 0)
				ct2 += responseHeaders.contentType.substring(idx);
			ln("Mapped '" + responseHeaders.contentType + "' to '" + ct2 + "'");
			Headers h1 = responseHeaders;
			Headers h2 = new Headers(h1.status, h1.statusMessage, ct2, h1.contentEncoding, h1.headerMap);
			rinfo.sendResultHeaders(h2);
		}
	}
}
