/* Copyright (c) 2006-2008, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.glimmerblock.proxy.request;

import java.io.IOException;
import com.sun.grizzly.tcp.OutputBuffer;
import com.sun.grizzly.tcp.Request;
import com.sun.grizzly.tcp.Response;
import com.sun.grizzly.util.buf.ByteChunk;
import com.sun.grizzly.util.http.MimeHeaders;
import dk.vitality.glimmerblock.api.result.Headers;
import dk.vitality.glimmerblock.api.result.HeadersBuilder;
import dk.vitality.glimmerblock.api.result.ResultData;
import dk.vitality.glimmerblock.api.util.Logger;

/**
 * Super-class of {@link RequestHandler} to avoid overly fat class.
 * <p/>
 * This class contains the information-part such as URL, HTTP headers, logging prefix
 * and a few utilities  while {@link RequestHandler} contains the processing methods.
 */
abstract public class RequestUtils implements Logger
{
	/**
	 * The Grizzly request object.
	 */
	public final Request request;

	/**
	 * The Grizzly response object.
	 */
	public final Response response;

	/**
	 * The HTTP request method (from the client) such as "GET".
	 */
	public final String requestMethod;

	/**
	 * If this is method "HEAD".
	 */
	public final boolean isHEAD;

	/**
	 * The HTTP headers sent by the client.
	 */
	public final Headers requestHeaders;

	/**
	 * The request serial id.
	 * Used in logging prefix to make it easy to see which requests which made which log lines.
	 */
	public final int id = nextHandlerid++;
	//
	private static int nextHandlerid = 100;

	/**
	 * The request-specific log prefix: "[" + id + "]"
	 */
	private final String logPrefix;

	/**
	 * Is the initial log line (containing HTTP method and request URL)
	 * for this request been emitted?
	 */
	private boolean hasMadeLogHeader;

	/**
	 * Proxy server logger which prefixes with date.
	 */
	private final Logger parentLogger;

	/**
	 * Wants debug logging?
	 */
	public boolean debugFlag = false;

	/**
	 * Wants verbose debug logging?
	 */
	public boolean debugVerbose = false;

	protected RequestUtils(Request request,
	                       Response response,
	                       Logger parentLogger,
	                       boolean debugFlag,
	                       boolean debugVerbose)
	{
		this.request = request;
		this.response = response;
		this.requestMethod = request.method().toString();
		this.isHEAD = requestMethod.equals("HEAD");
		this.parentLogger = parentLogger;
		this.logPrefix = "[" + id + "] ";
		this.debugFlag = debugFlag;
		this.debugVerbose = debugVerbose;
		HeadersBuilder hb = new HeadersBuilder();
		MimeHeaders mh = request.getMimeHeaders();
		for (int i = 0; i < mh.size(); i++) {
			String nameLC = mh.getName(i).toString().toLowerCase();
			String value = mh.getValue(i).toString();
			hb.add(nameLC, value);
		}
		requestHeaders = hb.toHeaders();
	}

	/**
	 * See {@link Logger#ln}
	 */
	public void ln(String s)
	{
		if (!hasMadeLogHeader) {
			hasMadeLogHeader = true;
			parentLogger.ln(logPrefix + requestMethod + "  " + request.unparsedURI());
		}
		if (s == null)
			return;
		parentLogger.ln(logPrefix + "  " + s);
	}

	/**
	 * Send the specified HTTP headers to the client.
	 */
	public void sendResultHeaders(Headers r) throws Exception
	{
		if (debugVerbose)
			ln("  sendResultHeaders: " + r.status + " " + r.statusMessage);
		response.setStatus(r.status);
		response.setMessage(r.statusMessage);
		if (r.contentType != null) {
			response.setHeader("Content-Type", r.contentType);
			if (debugVerbose)
				ln("    Content-Type: " + r.contentType);
		}
		if (r.contentEncoding != null) {
			response.setHeader("Content-Encoding", r.contentEncoding);
			if (debugVerbose)
				ln("    Content-Encoding: " + r.contentEncoding);
		}
		for (String nameLC : r.headerMap.keySet()) {
			String name = Headers.getMixedCaseHeaderName(nameLC);
			for (String value : r.headerMap.get(nameLC)) {
				response.addHeader(name, value);
				if (debugVerbose)
					ln("    " + name + ": " + value);
			}
		}
	}

	/**
	 * Send the specified HTTP headers and content to the client.
	 */
	public void sendResultData(ResultData rd) throws Exception
	{
		sendResultHeaders(rd.headers);
		ByteChunk chunk = rd.getContentAsByteChunk();
		if (chunk != null)
			response.doWrite(chunk);
	}

	/**
	 * Send simple HTTP error response (headers and content) to the client.
	 */
	public void sendError(int statusCode, String message) throws IOException
	{
		ln("SendError " + statusCode + ": " + message);
		response.setStatus(statusCode);
		byte[] bytes = message.getBytes("utf-8");
		ByteChunk chunk = new ByteChunk();
		response.setContentLength(bytes.length);
		response.setContentType("text/plain; charset=UTF-8; charset=UTF-8");
		response.setHeader("Pragma", "no-cache");
		response.setHeader("Cache-Control", "no-cache");
		chunk.append(bytes, 0, bytes.length);
		OutputBuffer buffer = response.getOutputBuffer();
		buffer.doWrite(chunk, response);
		response.finish();
	}
}