/* Copyright (c) 2006-2008, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.glimmerblock.proxy.urlExec;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.ConnectException;
import java.net.HttpURLConnection;
import java.net.Proxy;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.net.URL;
import static java.net.HttpURLConnection.HTTP_BAD_GATEWAY;
import static java.net.HttpURLConnection.HTTP_NOT_FOUND;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import dk.vitality.glimmerblock.api.result.Headers;
import dk.vitality.glimmerblock.api.result.HeadersBuilder;
import dk.vitality.glimmerblock.api.util.Logger;

/**
 * Base-class for a 'curl'-like class to get content from an HTTP server.
 * Sub-classes streams the content to the client or saves it in a byte array.
 * <p/>
 * It currently uses {@link java.net.HttpURLConnection} which isn't optimal.
 */
abstract public class AbstractUrlExec implements Logger
{
	/**
	 * The HTTP method, e.g. GET.
	 */
	public final String method;

	/**
	 * The requested URL.
	 */
	public final URL requestUrl;

	/**
	 * Final URL after handling redirects
	 */
	public URL responseUrl;

	/**
	 * The request headers, e.g. "User-Agent".
	 */
	public final Headers requestHeaders;

	/**
	 * Do we allow "Content-Encoding: gzip".  We can't just pass the header value from the client as
	 * subclasses must be able to decode the content from the server (to support modification of the content).
	 */
	public final boolean allowGzip;

	/**
	 * Should redirects from the server be handled automatically, or do we want to return the redirect
	 * headers to the client.
	 */
	public boolean followRedirects = false;

	/**
	 * Is debugging enabled (i.e. logging).
	 */
	public boolean debugFlag;

	/**
	 * The HTTP connection.  The java class library automatically handles keep-alive and connection pooling.
	 */
	protected HttpURLConnection conn;

	/**
	 * The HTTP headers sent by the server.
	 */
	public Headers responseHeaders;

	/**
	 * Number of redirects handled to avoid infinite loops from the server.
	 */
	protected int redirectCount;

	/**
	 * The proxy requests logger.
	 */
	private final Logger parentLogger;

	/**
	 * HTTP headers which always can be forwarded from the client, e.g. "User-Agent" and cookies.
	 */
	static public final Set<String> basicBrowserHeadersLC;

	static {
		Set<String> set = new HashSet<String>();
		set.add("user-agent");
		set.add("referer");
		set.add("accept");
		set.add("accept-language");
		set.add("if-modified-since");
		set.add("cookie");
		basicBrowserHeadersLC = Collections.unmodifiableSet(set);
	}

	/**
	 * HTTP headers which never must be forwarded, such as connection-handling headers (e.g. Keep-Alive).
	 * All headers whose name start with "Proxy" are ignored by {@link #copyBrowserHeaders} and is not
	 * included in this set.
	 */
	static public final Set<String> neverCopyHeadersLC;

	static {
		Set<String> set = new HashSet<String>();
		set.add("accept-encoding");
		set.add("keep-alive");
		set.add("connection");
		set.add("host");
		set.add("content-length");
		neverCopyHeadersLC = Collections.unmodifiableSet(set);
	}

	protected AbstractUrlExec(Logger parentLogger,
	                          String method,
	                          URL url,
	                          Headers requestHeaders,
	                          boolean allowGzip) throws Exception
	{
		this.parentLogger = parentLogger;
		this.method = method;
		this.requestUrl = url;
		this.responseUrl = url;
		this.requestHeaders = requestHeaders;
		this.allowGzip = allowGzip;
	}

	/**
	 * Main method for executing the request.
	 * Autmatically handles redirects if {@link #followRedirects} is true.
	 */
	public void execute() throws Exception
	{
		try {
			for (redirectCount = 0; true; redirectCount++) {
				checkRedirectCount();
				createConnectionObject();
				copyBrowserHeaders();
				processResponse();
				if (!isRedirect() || !followRedirects || !handleRedirect())
					return;
			}
		}
		finally {
			if (conn != null)
				conn.disconnect();
		}
	}

	/**
	 * Process the response from the server.
	 * Main function is to decode the nice variarity of Exceptions that the java library throws
	 * depending on IO errors or HTTP status codes.
	 */
	protected void processResponse() throws Exception
	{
		try {
			sendRequestBody();
			readResponseHeaders();
			readResponseBody();
		}
		catch (ConnectException ex) {
			Headers h = new Headers(HTTP_BAD_GATEWAY, "Upstream server denies connection.");
			throw new HttpException(h, ex);
		}
		catch (SocketTimeoutException ex) {
			Headers h = new Headers(HTTP_BAD_GATEWAY, "Upstream server doesn't respond at all.");
			throw new HttpException(h, ex);
		}
		catch (SocketException ex) {
			Headers h = new Headers(HTTP_BAD_GATEWAY, "Upstream server didn't send full HTTP response header.");
			throw new HttpException(h, ex);
		}
		catch (FileNotFoundException ex) {
			// Thrown when the server sends a 404 status.
			Headers h = new Headers(HTTP_NOT_FOUND, "Not found: " + requestUrl);
			throw new HttpException(h, ex);
		}
		catch (IOException ex) {
			Headers h = new Headers(HTTP_BAD_GATEWAY, "Got IOException (" + ex.getMessage() + "): " + requestUrl);
			throw new HttpException(h, ex);
		}
	}

	/**
	 * Create and configure the {@link HttpURLConnection} object.
	 */
	protected void createConnectionObject() throws Exception
	{
		if (conn != null)
			conn.disconnect();
		conn = (HttpURLConnection)responseUrl.openConnection(Proxy.NO_PROXY);
		conn.setAllowUserInteraction(false);
		conn.setConnectTimeout(10 * 1000);
		conn.setInstanceFollowRedirects(false);
		conn.setReadTimeout(600 * 1000);
		conn.setRequestMethod(method);
	}

	/**
	 * See {@link Logger#ln}
	 */
	public void ln(String s)
	{
		parentLogger.ln(s);
	}

	/**
	 * Copy the request headers from the client into the {@link HttpURLConnection} object.
	 */
	protected void copyBrowserHeaders() throws Exception
	{
		for (String nameLC : requestHeaders.headerMap.keySet()) {
			if (nameLC.startsWith("proxy") || neverCopyHeadersLC.contains(nameLC))
				continue;
			if (!shouldCopyBrowserHeader(nameLC))
				continue;
			String name = Headers.getMixedCaseHeaderName(nameLC);
			for (String value : requestHeaders.headerMap.get(nameLC)) {
				conn.setRequestProperty(name, value);
				if (debugFlag)
					ln("  Client-header: " + name + ": " + value);
			}
		}
		if (allowGzip)
			conn.setRequestProperty("Accept-Encoding", "gzip");
	}

	/**
	 * Should the specified header from the client be copied to the {@link HttpURLConnection} object.
	 *
	 * @param nameLC LOWERCASE name of the header.
	 */
	protected boolean shouldCopyBrowserHeader(String nameLC)
	{
		return basicBrowserHeadersLC.contains(nameLC);
	}

	/**
	 * Checks the number of redirects to avoid infinite loops from the server.
	 */
	protected void checkRedirectCount() throws Exception
	{
		if (redirectCount < 25)
			return;
		Headers h = new Headers(HTTP_BAD_GATEWAY, "Too many redirects from the server.",
				"text/plain", null, null);
		throw new HttpException(h);
	}

	/**
	 * Does the HTTP status from the server specify a redirect status (301, 302, 303).
	 */
	protected boolean isRedirect()
	{
		switch (responseHeaders.status) {
			case HttpURLConnection.HTTP_MOVED_PERM:
			case HttpURLConnection.HTTP_MOVED_TEMP:
			case HttpURLConnection.HTTP_SEE_OTHER:
				return true;
			default:
				return false;
		}
	}

	/**
	 * Handle a redirect, assuming both {@link #followRedirects} and {@link #isRedirect} are true.
	 * Updates {@link #responseUrl} with the new URL.
	 */
	protected boolean handleRedirect() throws Exception
	{
		String s = responseHeaders.headerMap.get("location").get(0);
		if (s == null)
			throw new Exception("Redirect without 'location' header");
		ln("Redirect to: " + s);
		responseUrl = new URL(responseUrl, s);
		if (s.startsWith("/"))
			ln("   absolute: " + responseUrl);
		return true;
	}

	/**
	 * Create {@link #responseHeaders} with the HTTP headers sent from the server.
	 */
	protected void readResponseHeaders() throws Exception
	{
		responseHeaders = readResponseHeaders(conn, this, false);
	}

	/**
	 * Read the response body from the server by e.g. storing it in a byte array
	 * or streaming it to the client.
	 */
	abstract protected void readResponseBody() throws Exception;

	/**
	 * Send the request body, e.g. the POST data.
	 */
	protected void sendRequestBody() throws Exception
	{
	}

	/**
	 * Utility function for reading HTTP response headers into a {@link Headers} object.
	 */
	public static Headers readResponseHeaders(HttpURLConnection conn,
	                                          Logger logger,
	                                          boolean debugFlag) throws IOException
	{
		HeadersBuilder hb = new HeadersBuilder(conn.getResponseCode(), conn.getResponseMessage());
		if (debugFlag)
			logger.ln(" Status: " + hb.status + " " + hb.statusMessage);
		Map<String, List<String>> serverHeaders = conn.getHeaderFields();
		for (String name : serverHeaders.keySet()) {
			if (name == null)
				continue; // http status line
			List<String> serverList = serverHeaders.get(name);
			String nameLC = name.toLowerCase(); // makes it possible to use get() on the map.
			boolean ignore;
			if (nameLC.equals("content-encoding")) {
				hb.contentEncoding = serverList.get(0);
				ignore = true;
			} else if (nameLC.equals("content-type")) {
				hb.contentType = serverList.get(0);
				ignore = true;
			} else {
				ignore = nameLC.equals("transfer-encoding")
						|| nameLC.equals("keep-alive")
						|| nameLC.equals("connection")
						|| nameLC.equals("content-length");
			}
			// The list has been reversed, i.e. the first value in the list
			// is the last header sent from the server.
			// Some sites (e.g. xtube) sends multiple Set-Cookie headers for the same cookie,
			// so the order is important.
			List<String> list = serverHeaders.get(name);
			for (int idx = list.size() - 1; idx >= 0; idx--) {
				String value = list.get(idx);
				if (debugFlag) {
					if (ignore)
						logger.ln(" Ignores (" + name + "): " + value);
					else
						logger.ln(" Response(" + name + "): " + value);
				}
				if (!ignore)
					hb.add(nameLC, value);
			}
		}
		return hb.toHeaders();
	}
}