/* Copyright (c) 2006-2008, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.glimmerblock.proxy.urlExec;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.zip.GZIPInputStream;
import dk.vitality.glimmerblock.api.result.Headers;
import dk.vitality.glimmerblock.api.result.ResultData;
import dk.vitality.glimmerblock.api.util.Logger;
import static dk.vitality.glimmerblock.proxy.util.ExceptionUtils.getExMsg;

/**
 * Obtains the content of an URL much like the 'curl' utility, except that
 * <ol>
 * <li>it copies usable headers from the request (e.g. User-Agent)</li>
 * <li>it allows for Content-Encoding:gzip and automatically decompresses the content.</li>
 * </ol>
 */
public class UrlContentGetter extends AbstractUrlExec
{
	protected byte[] responseContent;

	public UrlContentGetter(Logger parentLogger, String method, URL url, Headers requestHeaders, boolean allowGzip)
			throws Exception
	{
		super(parentLogger, method, url, requestHeaders, allowGzip);
		this.followRedirects = true;
	}

	public byte[] getResponseContent()
	{
		return responseContent;
	}

	public ResultData getResultData()
	{
		return new ResultData(responseHeaders, responseContent);
	}

	protected void readResponseBody() throws Exception
	{
		readResponseBodyIntoByteArray();
	}

	protected void readResponseBodyIntoByteArray() throws Exception
	{
		InputStream is;
		try {
			is = conn.getInputStream();
		}
		catch (FileNotFoundException ex) {
			throw new HttpException(responseHeaders, ex);
		}
		catch (Exception ex) {
			if (responseHeaders.status < 400) {
				ln("Failed opening connection: " + ex.getClass().getSimpleName() + " // " + ex.getMessage());
				ex.printStackTrace();
			}
			Headers h = new Headers(HttpURLConnection.HTTP_INTERNAL_ERROR,
					"Upstream server failed: " + getExMsg(ex),
					"text/plain", null, responseHeaders.headerMap);
			throw new HttpException(h, ex);
		}
		String value = conn.getHeaderField("Content-Length");
		int bufferSize = 4096;
		if (value != null)
			bufferSize = (int)Math.max(16, Math.min(16384, Long.parseLong(value)));
		byte[] buffer = new byte[bufferSize];
		ByteArrayOutputStream baos = new ByteArrayOutputStream(bufferSize);
		while (true) {
			int num = is.read(buffer);
			if (num < 0)
				break;
			baos.write(buffer, 0, num);
		}
		responseContent = baos.toByteArray();
		if (responseHeaders.contentEncoding != null && responseHeaders.contentEncoding.contains("gzip")) {
			//ln("    gzip decompress (" + allowGzip + "): " + responseContent.length + " : " + responseHeaders.contentEncoding);
			InputStream gz = new GZIPInputStream(new ByteArrayInputStream(responseContent));
			baos = new ByteArrayOutputStream(responseContent.length * 2);
			byte[] tmp = new byte[4096];
			int x;
			while ((x = gz.read(tmp)) >= 0)
				baos.write(tmp, 0, x);
			gz.close();
			//if (debugVerbose)
			ln("    gzip decompress: " + responseContent.length + " -> " + baos.size() + " bytes.");
			responseContent = baos.toByteArray();
		}
		responseHeaders = new Headers(responseHeaders.status, responseHeaders.statusMessage,
				responseHeaders.contentType, null, responseHeaders.headerMap);
	}
}
