/* Copyright (c) 2006-2008, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.glimmerblock.proxy.urlExec;

import java.io.IOException;
import java.io.InputStream;
import java.net.SocketTimeoutException;
import java.net.URL;
import com.sun.grizzly.tcp.ActionCode;
import com.sun.grizzly.util.buf.ByteChunk;
import dk.vitality.glimmerblock.api.result.Headers;
import dk.vitality.glimmerblock.api.util.Logger;
import dk.vitality.glimmerblock.proxy.request.ClientAbortedException;
import dk.vitality.glimmerblock.proxy.request.RequestUtils;
import static dk.vitality.glimmerblock.proxy.util.ExceptionUtils.getExMsg;

/**
 * Sends request to server and streams the result back to the client without modifications.
 * <p/>
 * It provides special handling for RTSP POST requests, see {@link #isRtspPOST}
 */
public class UrlStreamer extends UrlContentGetter
{
	protected final RequestUtils rinfo;

	public UrlStreamer(Logger parentLogger,
	                   String method,
	                   URL url,
	                   Headers requestHeaders,
	                   boolean allowGzip,
	                   RequestUtils rinfo) throws Exception
	{
		super(parentLogger, method, url, requestHeaders, allowGzip);
		this.rinfo = rinfo;
		this.followRedirects = false;
	}

	protected void readResponseHeaders() throws Exception
	{
		if (isRtspPOST())
			responseHeaders = new Headers("text/plain");
		else
			super.readResponseHeaders();
	}

	protected void readResponseBody() throws Exception
	{
		if (isRtspPOST())
			return;
		streamResponseBody();
	}

	/**
	 * Copy as many HTTP headers as possible as we intend to stream the result.
	 */
	protected boolean shouldCopyBrowserHeader(String nameLC)
	{
		return true;
	}

	protected void streamResponseBody() throws Exception
	{
		sendResultHeaders();
		sendOptContentLength();
		long st0 = System.currentTimeMillis();
		InputStream is = conn.getInputStream();
		long oldestData = System.currentTimeMillis();
		byte[] buffer = new byte[65536];
		ByteChunk chunk = new ByteChunk();
		while (true) {
			int num;
			long maxWait = 0;
			try {
				if (oldestData != 0) {
					long now = System.currentTimeMillis();
					maxWait = 100 - (now - oldestData);
					if (maxWait <= 0) {
						if (debugFlag)
							ln("!!! Flushes client data");
						rinfo.response.action(ActionCode.ACTION_CLIENT_FLUSH, rinfo.response);
						oldestData = 0;
					}
				}
				if (maxWait > 0) {
					if (debugFlag)
						ln("Using timeout = " + maxWait + " msecs.");
					conn.setReadTimeout((int)maxWait);
				} else {
					if (debugFlag)
						ln("Using 'blocking' read.");
					conn.setReadTimeout(5 * 60 * 1000);
				}
				if (debugFlag) {
					long st1 = System.currentTimeMillis();
					num = is.read(buffer, 0, buffer.length);
					long dur = System.currentTimeMillis() - st1;
					ln("Read time: " + dur + " msecs.");
				} else {
					num = is.read(buffer, 0, buffer.length);
				}
			}
			catch (SocketTimeoutException ex) {
				if (maxWait > 0)
					continue;
				if (debugFlag)
					ln("Got exception while using 'blocking' read from server: " + getExMsg(ex));
				break;
			}
			catch (IOException ex) {
				if (debugFlag)
					ln("Got exception when reading from server: " + getExMsg(ex));
				break;
			}
			if (num < 0) {
				if (debugFlag) {
					long dur = System.currentTimeMillis() - st0;
					ln("EOF of response data from server, total time = " + dur + " msecs.");
				}
				break;
			}
			if (num == 0)
				continue;
			if (debugFlag)
				ln("Forwards " + num + " bytes from server");
			chunk.setBytes(buffer, 0, num);
			try {
				rinfo.response.doWrite(chunk);
			}
			catch (IOException ex) {
				throw new ClientAbortedException("IOException: " + ex.getMessage(), ex);
			}
			if (oldestData == 0)
				oldestData = System.currentTimeMillis();
			//ln(new String(buffer, 0, num, "ISO-8859-1"));
		}
		is.close();
	}

	/**
	 * Forward the servers response HTTP headers to the client.
	 */
	protected void sendResultHeaders() throws Exception
	{
		rinfo.sendResultHeaders(responseHeaders);
	}

	/**
	 * Forward the Content-Length header from the servers response to the client.
	 */
	protected void sendOptContentLength()
	{
		String cl = conn.getHeaderField("content-length");
		if (cl != null && cl.length() > 0)
			rinfo.response.setContentLengthLong(Long.parseLong(cl));
	}

	/**
	 * <h2>How RTSP-over-HTTP works:</h2>
	 * (as observed by the behavior of QuickTime and Darwin Streaming Server)
	 * <p/>
	 * The client opens one GET request which is used to receive all data from the server.
	 * It keeps this connection during the full streaming session.
	 * The request must contain an "x-sessioncookie" header with a unique value. If the
	 * value is already known by the server (e.g. from an earlier request), the GET request fails
	 * without proper status code nor message.
	 * <p/>
	 * After the GET request has been sent, the client issues a POST request each time it wants
	 * to send a command to the server.  The request must contain the same "x-sessioncookie" header
	 * as the GET request.
	 * The RTSP command is base64 encoded and sent as the request-content.
	 * It does not use the standard URL-arg encoding.<br>
	 * When the server has received the POST content, it just keeps the connection open
	 * <i>and never sends a response nor closes the connection</i>. The client has thus to close
	 * the connection, very much against normal http pratice.
	 * <p/>
	 * A GET request looks like:
	 * <pre>
	 * GET http://streamer-01.dr.nordija.dk:80/dr1lowmp4 HTTP/1.0
	 * User-Agent: QuickTime/7.4.5 (qtver=7.4.5;cpu=IA32;os=Mac 10.5.2)
	 * x-sessioncookie: upYRxLgBAAD8fgkBPYcBAQ9
	 * Accept: application/x-rtsp-tunnelled
	 * Pragma: no-cache
	 * Cache-Control: no-cache
	 * </pre>
	 * and a POST request like:
	 * <pre>
	 * POST http://streamer-01.dr.nordija.dk:80/dr1lowmp4 HTTP/1.0
	 * User-Agent: QuickTime/7.4.5 (qtver=7.4.5;cpu=IA32;os=Mac 10.5.2)
	 * x-sessioncookie: upYRxLgBAAD8fgkBPYcBAQ9
	 * Content-Type: application/x-rtsp-tunnelled
	 * Pragma: no-cache
	 * Cache-Control: no-cache
	 * Content-Length: 32767
	 * Expires: Sun, 9 Jan 1972 00:00:00 GMT
	 * <b></b>
	 * REVTQ1JJQkUgcnRzcDovL3N0cmVhbWVyLTAxLmRyLm5vcmRpamEuZGs6ODAvZHIxbG93bXA0IFJUU1Av
	 * MS4wDQpDU2VxOiAxDQpBY2NlcHQ6IGFwcGxpY2F0aW9uL3NkcA0KQmFuZHdpZHRoOiAyMTQ3NDgzNjQ3
	 * DQpBY2NlcHQtTGFuZ3VhZ2U6IGVuLVVTDQpVc2VyLUFnZW50OiBRdWlja1RpbWUvNy40LjUgKHF0dmVy
	 * PTcuNC41O2NwdT1JQTMyO29zPU1hYyAxMC41LjIpDQoNCg==
	 * </pre>
	 * <a href='http://developer.apple.com/documentation/QuickTime/QTSS/Concepts/chapter_2_section_14.html'>Apple's documentation</a>
	 * says that the client might send several requests thru the same POST request. However, I've not seen
	 * this with QuickTime 7.4.5.
	 * The documentation too states that the content-length header of the request
	 * is bogous, so we can't depend on being able to read that many bytes from the client.
	 * <p/>
	 * So the GET conforms to standard http practice (assuming proper sessioncookie) while the POST doesn't.
	 */
	private boolean isRtspPOST()
	{
		return method.equals("POST") && "application/x-rtsp-tunnelled".equals(requestHeaders.contentType);
	}
}
