/* Copyright (c) 2006-2008, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.glimmerblock.proxy.util;

import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * Decodes an URL query-string into an hashtable and provides utility accessors for e.g. int params.
 * Empty parameters are ignored: in "a=42&b=&c=1" the "b" parameter is ignored and treated as if the
 * query string was "a=42&c=1".
 */
public class UrlParams implements Iterable<String>
{
	private Map<String, String[]> args = new HashMap<String, String[]>();

	/**
	 * Creates empty hash.
	 */
	public UrlParams()
	{
	}

	/**
	 * Copy-constructor. Makes deep copy.
	 */
	public UrlParams(UrlParams sp)
	{
		this.args.putAll(sp.args);
	}

	/**
	 * Gets an array of parameter values for non-empty parameter values.
	 *
	 * @return null if no such (non-empty) parameter was specified.
	 */
	public String[] optStringArray(String name)
	{
		if (name == null)
			throw new NullPointerException("name argument to get... may not be null");
		if (args == null)
			throw new NullPointerException("args hash is null ????");
		return args.get(name);
	}

	/**
	 * Is the named parameter specified and non-empty.
	 */
	public boolean has(String name)
	{
		return optStringArray(name) != null;
	}

	/**
	 * Get a parameter value and default to some value if the parameter was not specified (or was empty).
	 */
	public String optString(String name, String defaultValue)
	{
		String[] array = optStringArray(name);
		if (array == null || array.length == 0)
			return defaultValue;
		return array[0];
	}

	/**
	 * Get a parameter value and default to null if the parameter was not specified (or was empty).
	 */
	public String optString(String name)
	{
		return optString(name, null);
	}


	/**
	 * Get a required string parameter. Throws MissingServletArgumentException if not specified (or is empty).
	 */
	public String reqString(String name) throws MissingServletArgumentException
	{
		String s = optString(name);
		if (s == null)
			throw new MissingServletArgumentException(name);
		return s;
	}

	//-------------------------------------------------------------

	/**
	 * Get a required boolean parameter. Throws MissingServletArgumentException if not specified (or is empty).
	 */
	public boolean reqBoolean(String name) throws Exception
	{
		return convertStringToBoolean(reqString(name));
	}

	/**
	 * Get an optional boolean parameter. Returns false if not specified.
	 */
	public boolean optBoolean(String name) throws Exception
	{
		return optBoolean(name, false);
	}

	/**
	 * Get an optional boolean parameter. Returns the specified defaultValue if not specified.
	 */
	public boolean optBoolean(String name, boolean defaultValue) throws Exception
	{
		String s = optString(name);
		return (s == null) ? defaultValue : convertStringToBoolean(s);
	}

	/**
	 * Convert a string value into a boolean. Accepts "false" and "0" as false.
	 */
	private boolean convertStringToBoolean(String s) throws Exception
	{
		return !("0".equals(s) || "false".equalsIgnoreCase(s));
	}

	//-------------------------------------------------------------

	/**
	 * Similar to Collection.isEmpty: true if no parameters is specified.
	 */
	public boolean isEmpty()
	{
		return args.isEmpty();
	}

	/**
	 * Returns iterator for the parameter names.
	 */
	public Iterator<String> iterator()
	{
		return args.keySet().iterator();
	}

	/**
	 * Returns iterator for the parameter names.
	 */
	public Iterator<String> names()
	{
		return args.keySet().iterator();
	}

	//=========================================================================================
	/**
	 * Decode and add parameter (and values) from a query-string.
	 *
	 * @param qs
	 * @throws UnsupportedEncodingException
	 */
	public void addFromQueryString(String qs) throws UnsupportedEncodingException
	{
		addFromQueryString(qs, '&');
	}

	/**
	 * Decode and add parameter (and values) from a query-string.
	 *
	 * @param qs The parameter separator character, typically '&amp;'
	 */
	public void addFromQueryString(String qs, char sep) throws UnsupportedEncodingException
	{
		if (qs == null)
			return;
		while (qs.length() > 0) {
			int sepIdx = qs.indexOf(sep);
			String s = (sepIdx < 0) ? qs : qs.substring(0, sepIdx);
			qs = (sepIdx < 0) ? "" : qs.substring(sepIdx + 1);
			s = URLDecoder.decode(s, "ISO-8859-1");
			int eq = s.indexOf('=');
			if (eq < 0)
				continue;
			String name = s.substring(0, eq);
			while (name.startsWith("amp;"))
				name = name.substring(4); // multi-encoding of "&" -> "&amp;" in url
			String value = cleanString(s.substring(eq + 1));
			if (value.length() == 0)
				continue;
			//System.err.println("'" + name + "' -> ' " + value + "'");
			String[] oldArray = args.get(name);
			String[] newArray;
			if (oldArray == null) {
				newArray = new String[1];
				newArray[0] = value;
			} else {
				newArray = new String[oldArray.length + 1];
				System.arraycopy(oldArray, 0, newArray, 0, oldArray.length);
				newArray[oldArray.length] = value;
			}
			args.put(name, newArray);
		}

	}

	/**
	 * Clean the specified parameter value: the string is trim()ed and CRLF is changed into LF.
	 */
	private static String cleanString(String s)
	{
		s = s.replace('\u00A0', ' ');  // make nbsp into normal space so it is zapped by trim()
		if (s.indexOf('\r') < 0)
			return s.trim();
		for (int i = s.length() - 2; i >= 0; i--) {
			char ch = s.charAt(i);
			if (ch != '\r')
				continue;
			if (s.charAt(i + 1) != '\n')
				continue;
			s = i > 0 ? s.substring(0, i) + s.substring(i + 1) : s.substring(1);
		}
		s = s.replace('\r', '\n').trim();
		return s;
	}

	public static class MissingServletArgumentException extends Exception
	{
		public MissingServletArgumentException()
		{
		}

		public MissingServletArgumentException(String string)
		{
			super(string);
		}

		public MissingServletArgumentException(String string, Throwable throwable)
		{
			super(string, throwable);
		}

		public MissingServletArgumentException(Throwable throwable)
		{
			super(throwable);
		}
	}
}
